/* 
 * File:   Mp3_14.c
 *
 * Created on 2019/11/03
 */

#include <18F26K22.h>

#device ansi
#device const=read_only
#device PASS_STRINGS=IN_RAM
#opt compress

#fuses INTRC_IO,NOPLLEN,NOFCMEN,NOIESO,NOBROWNOUT,NOWDT,NOPBADEN,NOMCLR,STVREN,NOLVP,NOXINST,NOCPB,NOCPD,NODEBUG,CCP3B5
#use delay (clock=8000000)

#use rs232(uart1, baud=115200, bits=8, parity=N)        // デバッグ

#include <stdio.h>
#include <stdlib.h>

typedef unsigned int8   uint8_t;
typedef unsigned int16  uint16_t;
typedef unsigned int32  uint32_t;

#define TIMER0_INIT                 64536       // 15536
#define CHATTERING_ADJUST           200
#define CHECK_INTERVAL_TIME         2000UL    // IRパルス数カウントの区切り時間
#define MINIMUM_PULSE_TIMEWIDTH     50      // カウント対象にする最小パルス幅
#define POWERON_PULSE_COUNT         8
#define BLINK_TIME                  1000UL
#define TIMEOUT_ZERO_START          (1UL*60UL)         // Raspbery PIの起動待ちタイムアウト (秒)
#define TIMEOUT_ZERO_SHUTDOWN       (5UL*60UL*60UL)     // Raspbery PIの強制停止

#define PIN_DETECT_USB      PIN_A0
#define PIN_PLOCK           PIN_A1      // power lock (output)
#define PIN_POWER_SW        PIN_B5      // power sw (input)
#define PIN_DEVELOP         PIN_B6      // in develop 
#define PIN_ACTIVE_ZERO     PIN_A3      // (input)
#define PIN_IR_IN           PIN_B4      // IR sensor (input)
#define PIN_DIRECT_PDON     PIN_C1      // Raspbery PI へのPower Down指示 (output)
#define PIN_LED             PIN_C3

void InitApp();
void wait_ms(uint16_t ms);
void ExecRaspberryPi();

static uint16_t             WaitMsCycle;
static int                  LastIrStatus;
volatile static int         LowPulseCountReset;
volatile static uint16_t    TimerCount;
volatile static uint16_t    BlinkTimerCount;
volatile static int         BlinkStart;
static uint16_t             TimerLowStart;
static uint16_t             LowPulseCount;
static int                  LedStatus;
static volatile BOOLEAN     EVENT_POWERSW;
static uint16_t             CHATTERING_TMR;

int main(int argc, char** argv) {
    LowPulseCountReset = 0;
    TimerCount = 0;
    LowPulseCount = 0;
    BlinkTimerCount = 0;
    BlinkStart = 0;
    EVENT_POWERSW = 0;
    CHATTERING_TMR = 0;
    InitApp();
    LastIrStatus = input(PIN_IR_IN);

    if (input(PIN_DETECT_USB) == 1){    // USBバスパワー検出
        for (;;){
            printf("PIN_DETECT_USB == 1 \r\n");
            wait_ms(1000);          // Power On時のチャッタリングを飛ばす
            clear_interrupt(INT_RB4);
            enable_interrupts(INT_RB4);
            EVENT_POWERSW = 0;
            LowPulseCount = 0;
            CHATTERING_TMR = 0;

            // Raspberry PI 起動操作を待つ
            for (;;){
                if (LowPulseCount >= POWERON_PULSE_COUNT          // IRによる起動
                        || EVENT_POWERSW == 1               // パネルSWによる起動
                        || input(PIN_ACTIVE_ZERO) == 1){    // Raspberry PI すでに起動している
                    break;
                }
                if (LowPulseCountReset == 1){
                    //printf("LowPulseCount=%lu \r\n", LowPulseCount);
                    LowPulseCountReset = 0;
                    LowPulseCount = 0;
                }
            }
            disable_interrupts(INT_RB4);
            printf("zero start \r\n");
            ExecRaspberryPi();
            output_low(PIN_PLOCK);
        }
    } else {    // USBバスパワー無し
        output_high(PIN_PLOCK);
        wait_ms(1000);          // Power On時のチャッタリングを飛ばす
        printf("PIN_DETECT_USB == 0 \r\n");
        ExecRaspberryPi();
        disable_interrupts(GLOBAL);
        output_low(PIN_PLOCK);
        while (1){

        }
    }
    
}

void InitApp(){
    /* Initialize peripherals */
    setup_adc(ADC_OFF);         // ADC disabled
    setup_ccp1(CCP_OFF);        // CCP1 disabled
    setup_ccp2(CCP_OFF);        // CCP2 disabled
    setup_ccp3(CCP_OFF);        // CCP3 disabled
    
    setup_oscillator(OSC_8MHZ);
    
    // (1000us)/(4/8MHz)*250 = 8
    // -> 8*delay_cycle(250) = 1ms
    WaitMsCycle = 8;
        
    port_b_pullups(0x04 | 0x20);     // B6 B5 (01100000)
    output_float(PIN_DETECT_USB);
    output_float(PIN_IR_IN);
    output_high(PIN_DIRECT_PDON);
    output_float(PIN_ACTIVE_ZERO);
    output_float(PIN_POWER_SW);     // B5
    output_float(PIN_DEVELOP);      // B6
    output_low(PIN_PLOCK);
    
    // タイマーイベントハンドラが呼ばれる周期
    // Timer0 : (4/8MHz) * 2(prescaler) = 1us
    // 1us * 65536 = 65.5ms (TIMER0_INIT=0のとき)
    // 1us * (65536 - 15536) = 50.0ms (TIMER0_INIT=15536のとき)
    // 1us * (65536 - 64536) = 1000us = 1ms  (TIMER0_INIT=64536のとき)
    setup_timer_0(T0_INTERNAL | T0_DIV_2);
    set_timer0(TIMER0_INIT);
    clear_interrupt(INT_TIMER0);
    enable_interrupts(INT_TIMER0);
    //clear_interrupt(INT_RB4);         // INT_RB4は必要な時だけ有効化
    //enable_interrupts(INT_RB4);
    enable_interrupts(GLOBAL);

}

#INT_RB
void isr_rb4(){
    int irstatus;
    
    irstatus = input(PIN_IR_IN);
    if (irstatus == 1){     // HIGH
        if (LastIrStatus == 0){     // RISE
            LastIrStatus = 1;
            if (TimerCount > TimerLowStart){
                if (TimerCount - TimerLowStart > MINIMUM_PULSE_TIMEWIDTH){
                    LowPulseCount++;
                }
            }
        }
    
    } else {    // LOW
        if (LastIrStatus == 1){     // FALL
            LastIrStatus = 0;
            TimerLowStart = TimerCount;     // LOW PULSEの開始
        }
    }
}
    
// タイマーイベントハンドラ
#INT_TIMER0
void isr_timer0(){    
    TimerCount++;
    
    // IRパルスのカウントリセット
    if (TimerCount == CHECK_INTERVAL_TIME){
        TimerCount = 0;
        LowPulseCountReset = 1;
    }
    
    // LED点滅処理
    if (BlinkStart == 1){
        BlinkTimerCount++;
        if (BlinkTimerCount == BLINK_TIME){
            LedStatus = 1 - LedStatus;
            if (LedStatus == 1) output_high(PIN_LED);
            else output_low(PIN_LED);
            BlinkTimerCount = 0;
        }
    }
    
    // SW
    if (CHATTERING_TMR > 0){
        CHATTERING_TMR++;
        if (CHATTERING_TMR == CHATTERING_ADJUST) CHATTERING_TMR = 0;
    } else {
        // POWER SW
        if (input(PIN_POWER_SW) == 0 && EVENT_POWERSW == 0){         // POWER SW 押された
            EVENT_POWERSW = 1;
            CHATTERING_TMR = 1;
        }
    }
    
    set_timer0(TIMER0_INIT);
}

// Raspberry PI の起動～停止
void ExecRaspberryPi(){
    int         halt_flag = 0;
    uint16_t    timeout_counter = 0;
    
    EVENT_POWERSW = 0;
    CHATTERING_TMR = 0;
    output_high(PIN_PLOCK);
    wait_ms(1000);
    LedStatus = 1;
    output_high(PIN_LED);
    BlinkStart = 1;        // LED Blink 開始
    
    while (input(PIN_DEVELOP) == 0){
        
    }
    // Raspberry PI zero の起動を待つ
    while (input(PIN_ACTIVE_ZERO) == 0){
        wait_ms(100);
        timeout_counter++;
        if (timeout_counter == (TIMEOUT_ZERO_START*10)){
            goto end_proc;
            //break;
        }
    }
    
    // Raspberry PI 正常起動
    EVENT_POWERSW = 0;
    BlinkStart = 0;        // Blink 停止
    LedStatus = 1;
    output_high(PIN_LED);
    printf("Raspberry PI start \r\n");
    
    // Raspberry PI zero の停止を待つ
    timeout_counter = 0;
    while (input(PIN_ACTIVE_ZERO) == 1){
        wait_ms(1000);
        timeout_counter++;
        if (halt_flag == 0 && EVENT_POWERSW == 1){    // 全体のシャットダウンをPICから始める
            timeout_counter = 0;
//            output_low(PIN_DIRECT_PDON);      // PICからのRaspberry PI停止に対応しない (2019/12/08)
            halt_flag = 1;                      // そのかわりPIC停止までに待ち時間を置く
            printf("Power down Start \r\n");
        } else if (halt_flag == 1){
            if (timeout_counter == 60){       // 強制終了 (60秒)
                break;
            }
        } else if (timeout_counter >= TIMEOUT_ZERO_SHUTDOWN){
//            output_low(PIN_DIRECT_PDON);      // PICからのRaspberry PI停止に対応しない (2019/12/08)
            break;
        }
    }
    
end_proc:
    printf("end_proc \r\n");
    BlinkStart = 1;         // LED Blink 開始
    wait_ms(10000);         // 上のループではmp3_14_menuの停止しただけなのでもう少し待つ
    BlinkStart = 0;         // Blink 停止
    LedStatus = 0;
    output_low(PIN_LED);
}


void wait_ms(uint16_t ms){
    uint16_t    i;
    uint16_t    m;
    
    for (m=0; m<ms; m++){
        
        // 1ms
        for (i=0; i<WaitMsCycle; i++){
            delay_cycles(250);
        }
    }
}

